@@ -20,9 +20,17 @@
  */
 
 #include "config.h"
+
+#include <errno.h>
 #include <string.h>
+
+#include <glib/gstdio.h>
+
+#include <telepathy-glib/telepathy-glib.h>
+
 #include "mcd-account-manager-default.h"
 #include "mcd-debug.h"
+#include "mcd-misc.h"
 
 #define PLUGIN_NAME "default-gkeyfile"
 #define PLUGIN_PRIORITY MCP_ACCOUNT_STORAGE_PLUGIN_PRIO_DEFAULT
@@ -286,6 +294,57 @@ _get_secrets_from_keyring (const McpAccountStorage *self,
                 g_warning ("Unsupported value type for %s.%s", account, name);
             }
 
+          if (!tp_strdiff (param, "password"))
+            {
+              /* Empathy 3.0 was meant to migrate passwords from MC to
+               * itself, but it couldn't complete the migration by
+               * deleting the password from MC, because MC had several
+               * bugs that meant deleting passwords didn't work. To atone
+               * for our past sins, detect an incomplete migration and
+               * complete it. */
+              GnomeKeyringResult empathy_ok =
+                GNOME_KEYRING_RESULT_NO_KEYRING_DAEMON;
+              GnomeKeyringAttributeList *empathy_match =
+                gnome_keyring_attribute_list_new ();
+              GList *empathy_items = NULL;
+
+              gnome_keyring_attribute_list_append_string (empathy_match,
+                  "account-id", account);
+              gnome_keyring_attribute_list_append_string (empathy_match,
+                  "param-name", "password");
+
+              empathy_ok = gnome_keyring_find_items_sync (
+                  GNOME_KEYRING_ITEM_GENERIC_SECRET, empathy_match,
+                  &empathy_items);
+
+              if (empathy_ok == GNOME_KEYRING_RESULT_OK &&
+                  empathy_items != NULL)
+                {
+                  KeyringSetData *ksd = g_slice_new0 (KeyringSetData);
+
+                  DEBUG ("An Empathy 3.0 password migration wasn't finished "
+                      "due to fd.o #42088. Finishing it now by deleting the "
+                      "password for %s", account);
+
+                  ksd->account = g_strdup (account);
+                  ksd->name = g_strdup ("password");
+                  ksd->set = FALSE;
+
+                  gnome_keyring_delete_password (&keyring_schema,
+                      _keyring_set_cb, ksd, NULL,
+                      "account", account,
+                      "param", "password",
+                      NULL);
+                }
+
+              gnome_keyring_found_list_free (empathy_items);
+
+              /* behave as if it had already been deleted, i.e. we never
+               * actually found it... */
+              param = NULL;
+              value = NULL;
+            }
+
           if (param != NULL && value != NULL)
             {
               gchar *key = g_strdup_printf ("param-%s", param);
@@ -342,7 +401,7 @@ G_DEFINE_TYPE_WITH_CODE (McdAccountManagerDefault, mcd_account_manager_default,
         account_storage_iface_init));
 
 static gchar *
-get_account_conf_filename (void)
+get_old_filename (void)
 {
   const gchar *base;
 
@@ -360,11 +419,18 @@ get_account_conf_filename (void)
     return g_build_filename (base, "accounts.cfg", NULL);
 }
 
+static gchar *
+account_filename_in (const gchar *dir)
+{
+  return g_build_filename (dir, "telepathy", "mission-control", "accounts.cfg",
+      NULL);
+}
+
 static void
 mcd_account_manager_default_init (McdAccountManagerDefault *self)
 {
   DEBUG ("mcd_account_manager_default_init");
-  self->filename = get_account_conf_filename ();
+  self->filename = account_filename_in (g_get_user_data_dir ());
   self->keyfile = g_key_file_new ();
   self->secrets = g_key_file_new ();
   self->removed = g_key_file_new ();
@@ -380,25 +446,6 @@ mcd_account_manager_default_class_init (McdAccountManagerDefaultClass *cls)
   DEBUG ("mcd_account_manager_default_class_init");
 }
 
-static gboolean
-_have_config (McdAccountManagerDefault *self)
-{
-  DEBUG ("checking for %s", self->filename);
-  return g_file_test (self->filename, G_FILE_TEST_EXISTS);
-}
-
-static void
-_create_config (McdAccountManagerDefault *self)
-{
-  gchar *dir = g_path_get_dirname (self->filename);
-
-  DEBUG ("");
-  g_mkdir_with_parents (dir, 0700);
-  g_free (dir);
-  g_file_set_contents (self->filename, INITIAL_CONFIG, -1, NULL);
-  DEBUG ("created %s", self->filename);
-}
-
 /* We happen to know that the string MC gave us is "sufficiently escaped" to
  * put it in the keyfile as-is. */
 static gboolean
@@ -524,6 +571,25 @@ _get (const McpAccountStorage *self,
   return TRUE;
 }
 
+static gchar *
+_create (const McpAccountStorage *self,
+    const McpAccountManager *am,
+    const gchar *manager,
+    const gchar *protocol,
+    GHashTable *params,
+    GError **error)
+{
+  gchar *unique_name;
+
+  /* See comment in plugin-account.c::_storage_create_account() before changing
+   * this implementation, it's more subtle than it looks */
+  unique_name = mcp_account_manager_get_unique_name (MCP_ACCOUNT_MANAGER (am),
+                                                     manager, protocol, params);
+  g_return_val_if_fail (unique_name != NULL, NULL);
+
+  return unique_name;
+}
+
 static gboolean
 _delete (const McpAccountStorage *self,
       const McpAccountManager *am,
@@ -589,16 +655,39 @@ _commit (const McpAccountStorage *self,
   gchar *data;
   McdAccountManagerDefault *amd = MCD_ACCOUNT_MANAGER_DEFAULT (self);
   gboolean rval = FALSE;
+  gchar *dir;
+  GError *error = NULL;
 
   if (!amd->save)
     return TRUE;
 
-  if (!_have_config (amd))
-    _create_config (amd);
+  dir = g_path_get_dirname (amd->filename);
+
+  DEBUG ("Saving accounts to %s", amd->filename);
+
+  if (!mcd_ensure_directory (dir, &error))
+    {
+      g_warning ("%s", error->message);
+      g_error_free (error);
+      /* fall through anyway: writing to the file will fail, but it does
+       * give us a chance to commit to the keyring too */
+    }
+
+  g_free (dir);
 
   data = g_key_file_to_data (amd->keyfile, &n, NULL);
-  rval = g_file_set_contents (amd->filename, data, n, NULL);
-  amd->save = !rval;
+  rval = g_file_set_contents (amd->filename, data, n, &error);
+
+  if (rval)
+    {
+      amd->save = FALSE;
+    }
+  else
+    {
+      g_warning ("%s", error->message);
+      g_error_free (error);
+    }
+
   g_free (data);
 
   _keyring_commit (self, am, account);
@@ -606,6 +695,31 @@ _commit (const McpAccountStorage *self,
   return rval;
 }
 
+static void
+am_default_load_keyfile (McdAccountManagerDefault *self,
+    const gchar *filename)
+{
+  GError *error = NULL;
+
+  if (g_key_file_load_from_file (self->keyfile, filename,
+        G_KEY_FILE_KEEP_COMMENTS, &error))
+    {
+      DEBUG ("Loaded accounts from %s", filename);
+    }
+  else
+    {
+      DEBUG ("Failed to load accounts from %s: %s", filename, error->message);
+      g_error_free (error);
+
+      /* Start with a blank configuration, but do not save straight away;
+       * we don't want to overwrite a corrupt-but-maybe-recoverable
+       * configuration file with an empty one until given a reason to
+       * do so. */
+      g_key_file_load_from_data (self->keyfile, INITIAL_CONFIG, -1,
+          G_KEY_FILE_KEEP_COMMENTS, NULL);
+    }
+}
+
 static GList *
 _list (const McpAccountStorage *self,
     const McpAccountManager *am)
@@ -616,12 +730,72 @@ _list (const McpAccountStorage *self,
   GList *rval = NULL;
   McdAccountManagerDefault *amd = MCD_ACCOUNT_MANAGER_DEFAULT (self);
 
-  if (!_have_config (amd))
-    _create_config (amd);
+  if (!amd->loaded && g_file_test (amd->filename, G_FILE_TEST_EXISTS))
+    {
+      /* If the file exists, but loading it fails, we deliberately
+       * do not fall through to the "initial configuration" case,
+       * because we don't want to overwrite a corrupted file
+       * with an empty one until an actual write takes place. */
+      am_default_load_keyfile (amd, amd->filename);
+      amd->loaded = TRUE;
+    }
 
   if (!amd->loaded)
-    amd->loaded = g_key_file_load_from_file (amd->keyfile, amd->filename,
-        G_KEY_FILE_KEEP_COMMENTS, NULL);
+    {
+      const gchar * const *iter;
+
+      for (iter = g_get_system_data_dirs ();
+          iter != NULL && *iter != NULL;
+          iter++)
+        {
+          gchar *filename = account_filename_in (*iter);
+
+          if (g_file_test (filename, G_FILE_TEST_EXISTS))
+            {
+              am_default_load_keyfile (amd, filename);
+              amd->loaded = TRUE;
+              /* Do not set amd->save: we don't need to write it to a
+               * higher-priority directory until it actually changes. */
+            }
+
+          g_free (filename);
+
+          if (amd->loaded)
+            break;
+        }
+    }
+
+  if (!amd->loaded)
+    {
+      gchar *old_filename = get_old_filename ();
+
+      if (g_file_test (old_filename, G_FILE_TEST_EXISTS))
+        {
+          am_default_load_keyfile (amd, old_filename);
+          amd->loaded = TRUE;
+          amd->save = TRUE;
+
+          if (_commit (self, am, NULL))
+            {
+              DEBUG ("Migrated %s to new location: deleting old copy");
+              if (g_unlink (old_filename) != 0)
+                g_warning ("Unable to delete %s: %s", old_filename,
+                    g_strerror (errno));
+            }
+        }
+
+      g_free (old_filename);
+    }
+
+  if (!amd->loaded)
+    {
+      DEBUG ("Creating initial account data");
+      g_key_file_load_from_data (amd->keyfile, INITIAL_CONFIG, -1,
+          G_KEY_FILE_KEEP_COMMENTS, NULL);
+      amd->loaded = TRUE;
+      amd->save = TRUE;
+      _commit (self, am, NULL);
+    }
 
   accounts = g_key_file_get_groups (amd->keyfile, &n);
 
@@ -646,6 +820,7 @@ account_storage_iface_init (McpAccountStorageIface *iface,
 
   mcp_account_storage_iface_implement_get (iface, _get);
   mcp_account_storage_iface_implement_set (iface, _set);
+  mcp_account_storage_iface_implement_create (iface, _create);
   mcp_account_storage_iface_implement_delete (iface, _delete);
   mcp_account_storage_iface_implement_commit_one (iface, _commit);
   mcp_account_storage_iface_implement_list (iface, _list);
